/
usr
/
local
/
lib64
/
python3.6
/
site-packages
/
torch
/
include
/
ATen
/
native
/
cuda
/
/usr/local/lib64/python3.6/site-packages/torch/include/ATen/native/cuda
mkdir
upload
Name
Size
Mode
Actions
BatchLinearAlgebraLib.h
3114
0644
edit
dl
rm
block_reduce.cuh
2549
0644
edit
dl
rm
CompositeRandomAccessor.h
929
0644
edit
dl
rm
CUDALoops.cuh
7598
0644
edit
dl
rm
CuFFTPlanCache.h
19282
0644
edit
dl
rm
CuFFTUtils.h
1892
0644
edit
dl
rm
DeviceSqrt.cuh
585
0644
edit
dl
rm
DistributionTemplates.h
27435
0644
edit
dl
rm
EmbeddingBackwardKernel.cuh
715
0644
edit
dl
rm
ForeachFunctors.cuh
16851
0644
edit
dl
rm
GridSampler.cuh
11316
0644
edit
dl
rm
im2col.cuh
6577
0644
edit
dl
rm
KernelUtils.cuh
2553
0644
edit
dl
rm
LaunchUtils.h
306
0644
edit
dl
rm
Loops.cuh
9997
0644
edit
dl
rm
Math.cuh
13840
0644
edit
dl
rm
MemoryAccess.cuh
12463
0644
edit
dl
rm
MiscUtils.h
3341
0644
edit
dl
rm
MultiTensorApply.cuh
7552
0644
edit
dl
rm
Normalization.cuh
74441
0644
edit
dl
rm
PersistentSoftmax.cuh
14635
0644
edit
dl
rm
Randperm.cuh
2114
0644
edit
dl
rm
Reduce.cuh
38784
0644
edit
dl
rm
Resize.cuh
1919
0644
edit
dl
rm
ROCmLoops.cuh
13526
0644
edit
dl
rm
SortingCommon.cuh
5688
0644
edit
dl
rm
SortingRadixSelect.cuh
11918
0644
edit
dl
rm
SortUtils.cuh
5549
0644
edit
dl
rm
TensorModeKernel.cuh
14391
0644
edit
dl
rm
UniqueCub.cuh
345
0644
edit
dl
rm
UpSample.cuh
7552
0644
edit
dl
rm
vol2col.cuh
8297
0644
edit
dl
rm
Edit:
/usr/local/lib64/python3.6/site-packages/torch/include/ATen/native/cuda/CUDALoops.cuh
(7598B)
#pragma once // This file provides two functions to help write GPU elementwise kernels: // // gpu_kernel(TensorIterator iter, <lambda>) // gpu_kernel_with_scalars(TensorIterator iter, <lambda>) // // The gpu_kernel_with_scalars generates specializations that support a // single scalar CPU argument, such as from `cuda_tensor + 5`. The CPU scalar // is lifted to a kernel parameter instead of copying to device memory. // This should be used in conjunction with TensorIterator::allow_cpu_scalars_, // which is the default for TensorIterator::binary_op. Otherwise, all inputs // and the output must be on the GPU. // // For example, to write a reciprocal kernel for GPU float Tensors: // // gpu_kernel(iter, []GPU_LAMBDA(float a) { // return 1.0f / a; // }); // // To write a multiplication kernel for GPU float Tensors where one argument // may be a CPU scalar: // // gpu_kernel_with_scalars(iter, []GPU_LAMBDA(float a, float b) { // return a * b; // }); // // See BinaryOpsKernel.cu for the complete implementation // #include <type_traits> #include <tuple> #include <ATen/ATen.h> #include <ATen/cuda/CUDAContext.h> #include <ATen/core/Array.h> #include <ATen/detail/FunctionTraits.h> #include <ATen/native/TensorIterator.h> #include <c10/macros/Macros.h> #include <c10/core/ScalarType.h> #include <c10/util/TypeCast.h> #include <c10/util/C++17.h> // Marks a lambda as executable on both the host and device. The __host__ // attribute is important so that we can access static type information from // the host, even if the function is typically only executed on the device. #ifndef GPU_LAMBDA #define GPU_LAMBDA __host__ __device__ #endif #ifdef __NVCC__ #define ASSERT_HOST_DEVICE_LAMBDA(type) \ static_assert(__nv_is_extended_host_device_lambda_closure_type(type), \ #type " must be a __host__ __device__ lambda") #else #define ASSERT_HOST_DEVICE_LAMBDA(type) #endif namespace at { namespace native { template<int vec_size, typename func_t, typename array_t> C10_LAUNCH_BOUNDS_1(num_threads) __global__ void vectorized_elementwise_kernel(int N, func_t f, array_t data) { using traits = function_traits<func_t>; int remaining = N - block_work_size * blockIdx.x; if (remaining < block_work_size) { // if this block handles the reminder, just do a naive unrolled loop auto input_calc = TrivialOffsetCalculator<traits::arity>(); auto output_calc = TrivialOffsetCalculator<1>(); auto loader = memory::LoadWithoutCast(); auto storer = memory::StoreWithoutCast(); auto policy = memory::policies::unroll<array_t, decltype(input_calc), decltype(output_calc), memory::LoadWithoutCast, memory::StoreWithoutCast>( data, remaining, input_calc, output_calc, loader, storer); elementwise_kernel_helper(f, policy); } else { // if this block has a full `block_work_size` data to handle, use vectorized memory access elementwise_kernel_helper(f, memory::policies::vectorized<vec_size, array_t>(data)); } } template<typename func_t, typename array_t, typename inp_calc_t, typename out_calc_t, typename loader_t, typename storer_t> C10_LAUNCH_BOUNDS_1(num_threads) __global__ void unrolled_elementwise_kernel(int N, func_t f, array_t data, inp_calc_t ic, out_calc_t oc, loader_t l, storer_t s) { int remaining = N - block_work_size * blockIdx.x; auto policy = memory::policies::unroll<array_t, inp_calc_t, out_calc_t, loader_t, storer_t>(data, remaining, ic, oc, l, s); elementwise_kernel_helper(f, policy); } // this function assume trivial 1d and no dynamic casting template<typename func_t, typename array_t> static inline void launch_vectorized_kernel(int64_t N, const func_t& f, array_t data) { TORCH_INTERNAL_ASSERT(N > 0 && N <= std::numeric_limits<int32_t>::max()); using traits = function_traits<func_t>; int64_t grid = (N + block_work_size - 1) / block_work_size; auto stream = at::cuda::getCurrentCUDAStream(); int vec_size = memory::can_vectorize_up_to<func_t>(data); switch (vec_size) { case 4: vectorized_elementwise_kernel<4, func_t, array_t><<<grid, num_threads, 0, stream>>>(N, f, data); C10_CUDA_KERNEL_LAUNCH_CHECK(); break; case 2: vectorized_elementwise_kernel<2, func_t, array_t><<<grid, num_threads, 0, stream>>>(N, f, data); C10_CUDA_KERNEL_LAUNCH_CHECK(); break; case 1: { auto input_calc = TrivialOffsetCalculator<traits::arity>(); auto output_calc = TrivialOffsetCalculator<1>(); auto loader = memory::LoadWithoutCast(); auto storer = memory::StoreWithoutCast(); unrolled_elementwise_kernel<func_t, array_t><<<grid, num_threads, 0, stream>>>(N, f, data, input_calc, output_calc, loader, storer); C10_CUDA_KERNEL_LAUNCH_CHECK(); break; } default: TORCH_INTERNAL_ASSERT(false, "Unexpected vectorization size"); } } template<typename func_t, typename array_t, typename inp_calc_t, typename out_calc_t, typename loader_t, typename storer_t> static inline void launch_unrolled_kernel(int64_t N, const func_t& f, array_t data, inp_calc_t ic, out_calc_t oc, loader_t l, storer_t s) { TORCH_INTERNAL_ASSERT(N > 0 && N <= std::numeric_limits<int32_t>::max()); int64_t grid = (N + block_work_size - 1) / block_work_size; auto stream = at::cuda::getCurrentCUDAStream(); unrolled_elementwise_kernel<func_t, array_t><<<grid, num_threads, 0, stream>>>(N, f, data, ic, oc, l, s); C10_CUDA_KERNEL_LAUNCH_CHECK(); } template <typename func_t> void gpu_kernel_impl(TensorIteratorBase& iter, const func_t& f) { using traits = function_traits<func_t>; using arg0_t = typename traits::result_type; constexpr int ntensors = traits::arity + 1; TORCH_INTERNAL_ASSERT(iter.can_use_32bit_indexing()); TORCH_INTERNAL_ASSERT(iter.ninputs() == traits::arity); TORCH_INTERNAL_ASSERT(iter.noutputs() == 1); at::detail::Array<char*, ntensors> data; for (int i = 0; i < ntensors; i++) { data[i] = (char*)iter.data_ptr(i); } int64_t numel = iter.numel(); bool contiguous = iter.is_contiguous(); bool dynamic_casting = needs_dynamic_casting<func_t>::check(iter); if (!dynamic_casting) { if (contiguous) { launch_vectorized_kernel(numel, f, data); } else { auto input_offset_calculator = make_input_offset_calculator<traits::arity>(iter); auto output_offset_calculator = make_output_offset_calculator(iter); auto loader = memory::LoadWithoutCast(); auto storer = memory::StoreWithoutCast(); launch_unrolled_kernel(numel, f, data, input_offset_calculator, output_offset_calculator, loader, storer); } } else { at::detail::Array<ScalarType, traits::arity> dtypes; for (int i = 0; i < traits::arity; i++) { dtypes[i] = iter.tensor(i + 1).scalar_type(); } auto loader = memory::LoadWithCast<traits::arity>(dtypes); auto storer = memory::StoreWithCast(iter.tensor(0).scalar_type()); if (contiguous) { auto input_offset_calculator = TrivialOffsetCalculator<traits::arity>(); auto output_offset_calculator = TrivialOffsetCalculator<1>(); launch_unrolled_kernel(numel, f, data, input_offset_calculator, output_offset_calculator, loader, storer); } else { auto input_offset_calculator = make_input_offset_calculator<traits::arity>(iter); auto output_offset_calculator = make_output_offset_calculator(iter); launch_unrolled_kernel(numel, f, data, input_offset_calculator, output_offset_calculator, loader, storer); } } } }} // namespace at::native
Save
cmd:
run